@@ -7,11 +7,15 @@ import email
 import unittest
 import time
 import re
+import mock
+from contextlib import contextmanager
 
 from email.header import Header
+from email import charset
 
 from flask import Flask
-from flask_mail import Mail, Message, BadHeaderError, sanitize_address
+from flask_mail import Mail, Message, BadHeaderError, sanitize_address, PY3
+from speaklater import make_lazy_string
 
 
 class TestCase(unittest.TestCase):
@@ -30,6 +34,24 @@ class TestCase(unittest.TestCase):
     def tearDown(self):
         self.ctx.pop()
 
+    @contextmanager
+    def mail_config(self, **settings):
+        """
+        Context manager to alter mail config during a test and restore it after,
+        even in case of a failure.
+        """
+        original = {}
+        state = self.mail.state
+        for key in settings:
+            assert hasattr(state, key)
+            original[key] = getattr(state, key)
+            setattr(state, key, settings[key])
+
+        yield
+        # restore
+        for k, v in original.items():
+            setattr(state, k, v)
+
     def assertIn(self, member, container, msg=None):
         if hasattr(unittest.TestCase, 'assertIn'):
             return unittest.TestCase.assertIn(self, member, container, msg)
@@ -51,6 +73,18 @@ class TestCase(unittest.TestCase):
         return self.assertTrue(obj is not None)
 
 
+class TestInitialization(TestCase):
+
+    def test_init_mail(self):
+        mail = self.mail.init_mail(
+            self.app.config,
+            self.app.debug,
+            self.app.testing
+        )
+
+        self.assertEquals(self.mail.state.__dict__, mail.__dict__)
+
+
 class TestMessage(TestCase):
 
     def test_initialize(self):
@@ -66,6 +100,17 @@ class TestMessage(TestCase):
         msg2.add_recipient("somebody@here.com")
         self.assertEqual(len(msg2.recipients), 1)
 
+    def test_esmtp_options_properly_initialized(self):
+        msg = Message(subject="subject")
+        self.assertEqual(msg.mail_options, [])
+        self.assertEqual(msg.rcpt_options, [])
+
+        msg = Message(subject="subject", mail_options=['BODY=8BITMIME'])
+        self.assertEqual(msg.mail_options, ['BODY=8BITMIME'])
+
+        msg2 = Message(subject="subject", rcpt_options=['NOTIFY=SUCCESS'])
+        self.assertEqual(msg2.rcpt_options, ['NOTIFY=SUCCESS'])
+
     def test_sendto_properly_set(self):
         msg = Message(subject="subject", recipients=["somebody@here.com"],
                       cc=["cc@example.com"], bcc=["bcc@example.com"])
@@ -83,6 +128,11 @@ class TestMessage(TestCase):
                       sender=("tester", "tester@example.com"))
         self.assertEqual('tester <tester@example.com>', msg.sender)
 
+    def test_default_sender_as_tuple(self):
+        self.app.extensions['mail'].default_sender = ('tester', 'tester@example.com')
+        msg = Message(subject="testing")
+        self.assertEqual('tester <tester@example.com>', msg.sender)
+
     def test_reply_to(self):
         msg = Message(subject="testing",
                       recipients=["to@example.com"],
@@ -136,7 +186,36 @@ class TestMessage(TestCase):
         self.assertEqual(a.data, b"this is a test")
 
     def test_bad_header_subject(self):
-        msg = Message(subject="testing\n\r",
+        msg = Message(subject="testing\r\n",
+                      sender="from@example.com",
+                      body="testing",
+                      recipients=["to@example.com"])
+        self.assertRaises(BadHeaderError, self.mail.send, msg)
+
+    def test_multiline_subject(self):
+        msg = Message(subject="testing\r\n testing\r\n testing \r\n \ttesting",
+                      sender="from@example.com",
+                      body="testing",
+                      recipients=["to@example.com"])
+        self.mail.send(msg)
+        response = msg.as_string()
+        self.assertIn("From: from@example.com", str(response))
+        self.assertIn("testing\r\n testing\r\n testing \r\n \ttesting", str(response))
+
+    def test_bad_multiline_subject(self):
+        msg = Message(subject="testing\r\n testing\r\n ",
+                      sender="from@example.com",
+                      body="testing",
+                      recipients=["to@example.com"])
+        self.assertRaises(BadHeaderError, self.mail.send, msg)
+
+        msg = Message(subject="testing\r\n testing\r\n\t",
+                      sender="from@example.com",
+                      body="testing",
+                      recipients=["to@example.com"])
+        self.assertRaises(BadHeaderError, self.mail.send, msg)
+
+        msg = Message(subject="testing\r\n testing\r\n\n",
                       sender="from@example.com",
                       body="testing",
                       recipients=["to@example.com"])
@@ -144,7 +223,7 @@ class TestMessage(TestCase):
 
     def test_bad_header_sender(self):
         msg = Message(subject="testing",
-                      sender="from@example.com\n\r",
+                      sender="from@example.com\r\n",
                       recipients=["to@example.com"],
                       body="testing")
 
@@ -153,7 +232,7 @@ class TestMessage(TestCase):
     def test_bad_header_reply_to(self):
         msg = Message(subject="testing",
                       sender="from@example.com",
-                      reply_to="evil@example.com\n\r",
+                      reply_to="evil@example.com\r",
                       recipients=["to@example.com"],
                       body="testing")
 
@@ -169,7 +248,7 @@ class TestMessage(TestCase):
                           "to\r\n@example.com"],
                       body="testing")
 
-        self.assertIn('To: to@example.com\n', msg.as_string())
+        self.assertIn('To: to@example.com', msg.as_string())
 
     def test_emails_are_sanitized(self):
         msg = Message(subject="testing",
@@ -207,6 +286,48 @@ class TestMessage(TestCase):
 
         self.assertIn('Content-Type: multipart/mixed', msg.as_string())
 
+    def test_plain_message_with_ascii_attachment(self):
+        msg = Message(subject="subject",
+                      recipients=["to@example.com"],
+                      body="hello")
+
+        msg.attach(data=b"this is a test",
+                   content_type="text/plain",
+                   filename='test doc.txt')
+
+        self.assertIn('Content-Disposition: attachment; filename="test doc.txt"', msg.as_string())
+
+    def test_plain_message_with_unicode_attachment(self):
+        msg = Message(subject="subject",
+                      recipients=["to@example.com"],
+                      body="hello")
+
+        msg.attach(data=b"this is a test",
+                   content_type="text/plain",
+                   filename=u'ünicöde ←→ ✓.txt')
+
+        parsed = email.message_from_string(msg.as_string())
+
+        self.assertIn(re.sub(r'\s+', ' ', parsed.get_payload()[1].get('Content-Disposition')), [
+            'attachment; filename*="UTF8\'\'%C3%BCnic%C3%B6de%20%E2%86%90%E2%86%92%20%E2%9C%93.txt"',
+            'attachment; filename*=UTF8\'\'%C3%BCnic%C3%B6de%20%E2%86%90%E2%86%92%20%E2%9C%93.txt'
+            ])
+
+    def test_plain_message_with_ascii_converted_attachment(self):
+        with self.mail_config(ascii_attachments=True):
+            msg = Message(subject="subject",
+                          recipients=["to@example.com"],
+                          body="hello")
+
+            msg.attach(data=b"this is a test",
+                       content_type="text/plain",
+                       filename=u'ünicödeß ←.→ ✓.txt')
+
+            parsed = email.message_from_string(msg.as_string())
+            self.assertIn(
+                'Content-Disposition: attachment; filename="unicode . .txt"',
+                msg.as_string())
+
     def test_html_message(self):
         html_text = "<p>Hello World</p>"
         msg = Message(sender="from@example.com",
@@ -303,6 +424,12 @@ class TestMessage(TestCase):
         self.assertIn(h2.encode(), response)
         self.assertIn(h3.encode(), response)
 
+    def test_unicode_subject(self):
+        msg = Message(subject=make_lazy_string(lambda a: a, u"sübject"),
+                      sender='from@example.com',
+                      recipients=["to@example.com"])
+        self.assertIn('=?utf-8?q?s=C3=BCbject?=', msg.as_string())
+
     def test_extra_headers(self):
         msg = Message(sender="from@example.com",
                       subject="subject",
@@ -353,6 +480,62 @@ class TestMessage(TestCase):
         msg.attach(data=b"foobar", content_type='text/csv')
         self.assertIn('Content-Type: text/plain; charset="utf-8"', msg.as_string())
 
+        # unicode sender as tuple
+        msg = Message(sender=(u"送信者", "from@example.com"),
+                      subject=u"表題",
+                      recipients=["foo@bar.com"],
+                      reply_to=u"返信先 <somebody@example.com>",
+                      charset='shift_jis')  # japanese
+        msg.body = u'内容'
+        self.assertIn('From: =?iso-2022-jp?', msg.as_string())
+        self.assertNotIn('From: =?utf-8?', msg.as_string())
+        self.assertIn('Subject: =?iso-2022-jp?', msg.as_string())
+        self.assertNotIn('Subject: =?utf-8?', msg.as_string())
+        self.assertIn('Reply-To: =?iso-2022-jp?', msg.as_string())
+        self.assertNotIn('Reply-To: =?utf-8?', msg.as_string())
+        self.assertIn('Content-Type: text/plain; charset="iso-2022-jp"', msg.as_string())
+
+        # unicode subject sjis
+        msg = Message(sender="from@example.com",
+                      subject=u"表題",
+                      recipients=["foo@bar.com"],
+                      charset='shift_jis')  # japanese
+        msg.body = u'内容'
+        self.assertIn('Subject: =?iso-2022-jp?', msg.as_string())
+        self.assertIn('Content-Type: text/plain; charset="iso-2022-jp"', msg.as_string())
+
+        # unicode subject utf-8
+        msg = Message(sender="from@example.com",
+                      subject="subject",
+                      recipients=["foo@bar.com"],
+                      charset='utf-8')
+        msg.body = u'内容'
+        self.assertIn('Subject: subject', msg.as_string())
+        self.assertIn('Content-Type: text/plain; charset="utf-8"', msg.as_string())
+
+        # ascii subject
+        msg = Message(sender="from@example.com",
+                      subject="subject",
+                      recipients=["foo@bar.com"],
+                      charset='us-ascii')
+        msg.body = "normal ascii text"
+        self.assertNotIn('Subject: =?us-ascii?', msg.as_string())
+        self.assertIn('Content-Type: text/plain; charset="us-ascii"', msg.as_string())
+
+        # default charset
+        msg = Message(sender="from@example.com",
+                      subject="subject",
+                      recipients=["foo@bar.com"])
+        msg.body = "normal ascii text"
+        self.assertNotIn('Subject: =?', msg.as_string())
+        self.assertIn('Content-Type: text/plain; charset="utf-8"', msg.as_string())
+
+    def test_empty_subject_header(self):
+        msg = Message(sender="from@example.com",
+                      recipients=["foo@bar.com"])
+        msg.body = "normal ascii text"
+        self.mail.send(msg)
+        self.assertNotIn('Subject:', msg.as_string())
 
 class TestMail(TestCase):
 
@@ -439,3 +622,72 @@ class TestConnection(TestCase):
                       recipients=["to@example.com"])
         with self.mail.connect() as conn:
             self.assertRaises(BadHeaderError, conn.send, msg)
+
+    def test_sendmail_with_ascii_recipient(self):
+        with self.mail.connect() as conn:
+            with mock.patch.object(conn, 'host') as host:
+                msg = Message(subject="testing",
+                              sender="from@example.com",
+                              recipients=["to@example.com"],
+                              body="testing")
+                conn.send(msg)
+
+                host.sendmail.assert_called_once_with(
+                    "from@example.com",
+                    ["to@example.com"],
+                    msg.as_bytes() if PY3 else msg.as_string(),
+                    msg.mail_options,
+                    msg.rcpt_options
+                )
+
+    def test_sendmail_with_non_ascii_recipient(self):
+        with self.mail.connect() as conn:
+            with mock.patch.object(conn, 'host') as host:
+                msg = Message(subject="testing",
+                              sender="from@example.com",
+                              recipients=[u'ÄÜÖ → ✓ <to@example.com>'],
+                              body="testing")
+                conn.send(msg)
+
+                host.sendmail.assert_called_once_with(
+                    "from@example.com",
+                    ["=?utf-8?b?w4TDnMOWIOKGkiDinJM=?= <to@example.com>"],
+                    msg.as_bytes() if PY3 else msg.as_string(),
+                    msg.mail_options,
+                    msg.rcpt_options
+                )
+
+    def test_sendmail_with_ascii_body(self):
+        with self.mail.connect() as conn:
+            with mock.patch.object(conn, 'host') as host:
+                msg = Message(subject="testing",
+                              sender="from@example.com",
+                              recipients=["to@example.com"],
+                              body="body")
+                conn.send(msg)
+
+                host.sendmail.assert_called_once_with(
+                    "from@example.com",
+                    ["to@example.com"],
+                    msg.as_bytes() if PY3 else msg.as_string(),
+                    msg.mail_options,
+                    msg.rcpt_options
+                )
+
+    def test_sendmail_with_non_ascii_body(self):
+        with self.mail.connect() as conn:
+            with mock.patch.object(conn, 'host') as host:
+                msg = Message(subject="testing",
+                              sender="from@example.com",
+                              recipients=["to@example.com"],
+                              body=u"Öö")
+
+                conn.send(msg)
+
+                host.sendmail.assert_called_once_with(
+                    "from@example.com",
+                    ["to@example.com"],
+                    msg.as_bytes() if PY3 else msg.as_string(),
+                    msg.mail_options,
+                    msg.rcpt_options
+                )
